/* ---------------------------------------------------------------------
*
*  -- ScaLAPACK routine (version 1.0) --
*     University of Tennessee, Knoxville, Oak Ridge National Laboratory,
*     and University of California, Berkeley.
*     November 17, 1996
*
*  ---------------------------------------------------------------------
*/
/*
*  Include files
*/
#include "pblas.h"

void pcherk_( uplo, trans, n, k, alpha, A, ia, ja, desc_A, beta,
              C, ic, jc, desc_C )
/*
*  .. Scalar Arguments ..
*/
   F_CHAR      trans, uplo;
   int         * ia, * ic, * ja, * jc, * k, * n;
   float       * alpha, * beta;
/* ..
*  .. Array Arguments ..
*/
   int         desc_A[], desc_C[];
   complex     A[], C[];
{
/*
*  Purpose
*  =======
*
*  PCHERK  performs one of the hermitian rank k operations
*
*     sub( C ) := alpha*sub( A )*conjg( sub( A )' ) + beta*sub( C ),
*
*  or
*
*     sub( C ) := alpha*conjg( sub( A )' )*sub( A ) + beta*sub( C ),
*
*  where sub( C ) denotes C(IC:IC+N-1,JC:JC+N-1),
*
*        sub( A ) denotes A(IA:IA+N-1,JA:JA+K-1)  if TRANS = 'N',
*                         A(IA:IA+K-1,JA:JA+N-1)  otherwise.
*
*  Alpha and beta are real scalars, sub( C ) is an N-by-N hermitian
*  distributed matrix and sub( A ) is an N-by-K distributed matrix in
*  the first case and a K-by-N distributed matrix in the second case.
*
*  Notes
*  =====
*
*  Each global data object is described by an associated description
*  vector.  This vector stores the information required to establish
*  the mapping between an object element and its corresponding process
*  and memory location.
*
*  Let A be a generic term for any 2D block cyclicly distributed array.
*  Such a global array has an associated description vector descA.
*  In the following comments, the character _ should be read as
*  "of the global array".
*
*  NOTATION        STORED IN      EXPLANATION
*  --------------- -------------- --------------------------------------
*  DT_A   (global) descA[ DT_ ]   The descriptor type.  In this case,
*                                 DT_A = 1.
*  CTXT_A (global) descA[ CTXT_ ] The BLACS context handle, indicating
*                                 the BLACS process grid A is distribu-
*                                 ted over. The context itself is glo-
*                                 bal, but the handle (the integer
*                                 value) may vary.
*  M_A    (global) descA[ M_ ]    The number of rows in the global
*                                 array A.
*  N_A    (global) descA[ N_ ]    The number of columns in the global
*                                 array A.
*  MB_A   (global) descA[ MB_ ]   The blocking factor used to distribu-
*                                 te the rows of the array.
*  NB_A   (global) descA[ NB_ ]   The blocking factor used to distribu-
*                                 te the columns of the array.
*  RSRC_A (global) descA[ RSRC_ ] The process row over which the first
*                                 row of the array A is distributed.
*  CSRC_A (global) descA[ CSRC_ ] The process column over which the
*                                 first column of the array A is
*                                 distributed.
*  LLD_A  (local)  descA[ LLD_ ]  The leading dimension of the local
*                                 array.  LLD_A >= MAX(1,LOCr(M_A)).
*
*  Let K be the number of rows or columns of a distributed matrix,
*  and assume that its process grid has dimension p x q.
*  LOCr( K ) denotes the number of elements of K that a process
*  would receive if K were distributed over the p processes of its
*  process column.
*  Similarly, LOCc( K ) denotes the number of elements of K that a
*  process would receive if K were distributed over the q processes of
*  its process row.
*  The values of LOCr() and LOCc() may be determined via a call to the
*  ScaLAPACK tool function, NUMROC:
*          LOCr( M ) = NUMROC( M, MB_A, MYROW, RSRC_A, NPROW ),
*          LOCc( N ) = NUMROC( N, NB_A, MYCOL, CSRC_A, NPCOL ).
*  An upper bound for these quantities may be computed by:
*          LOCr( M ) <= ceil( ceil(M/MB_A)/NPROW )*MB_A
*          LOCc( N ) <= ceil( ceil(N/NB_A)/NPCOL )*NB_A
*
*  according to a square block cyclic decomposition, i.e MB_C = NB_C, if
*  N+MOD(IC-1,MB_C) > MB_C or N+MOD(JC-1,NB_C) > NB_C, in which case
*  sub( C ) is not just contained into a block and IC-1 (resp. JC-1) must
*  be a multiple of MB_C (resp. NB_C).
*
*  If TRANS = 'N', then sub( A ) and sub( C ) must be row aligned, i.e
*  the row process having the first entries of sub( A ) must also own
*  the first entries of sub( C ).
*  If sub( C ) is not just contained into a block, IC-1 (resp. IA-1)
*  must be a multiple of MB_C (resp. MB_A) and the column block size of
*  C must be equal to the row block size of A, i.e NB_C = MB_A.
*
*  Otherwise sub( A ) and sub( C ) must be column aligned, i.e the
*  column process having the first entries of sub( A ) must also own the
*  first entries of sub( C ).
*  If sub( C ) is not just contained into a block, JC-1 (resp. JA-1)
*  must be a multiple of NB_C (resp. NB_A) and the row block size of C
*  must be equal to the column block size of A, i.e MB_C = NB_A.
*
*  Parameters
*  ==========
*
*  UPLO    (global input) pointer to CHARACTER
*          On  entry,   UPLO  specifies  whether  the  upper  or  lower
*          triangular  part  of the  distributed matrix sub( C ) is to
*          be  referenced  as  follows:
*
*          UPLO = 'U' or 'u' Only the upper triangular part of sub( C )
*                            is to be referenced,
*
*          UPLO = 'L' or 'l' Only the lower triangular part of sub( C )
*                            is to be referenced.
*
*  TRANS   (global input) pointer to CHARACTER
*          On entry,  TRANS  specifies the operation to be performed as
*          follows:
*
*          TRANS = 'N' or 'n'
*                 sub( C ) := alpha*sub( A )*conjg( sub( A )' ) +
*                             beta*sub( C ),
*
*          TRANS = 'C' or 'c'
*                 sub( C ) := alpha*conjg( sub( A )' )*sub( A ) +
*                             beta*sub( C ).
*
*  N       (global input) pointer to INTEGER
*          The order of the distributed matrix sub( C ). N >= 0.
*
*  K       (global input) pointer to INTEGER
*          On entry with  TRANS = 'N' or 'n', K specifies the number of
*          columns  of the distributed matrix sub( A ), and on entry
*          with TRANS = 'C' or 'c', K specifies the number of rows of
*          the distributed matrix sub( A ). K >= 0.
*
*  ALPHA   (global input) pointer to REAL
*          On entry, ALPHA specifies the scalar alpha.
*
*  A       (local input) COMPLEX pointer into the local memory
*          to an array of dimension (LLD_A, KLa), where KLa is
*          LOCc(JA+K-1) when  TRANS = 'N' or 'n',  and is LOCc(JA+N-1)
*          otherwise.  Before entry with TRANS = 'N' or 'n', this array
*          contains the local pieces of the distributed matrix sub( A ).
*
*  IA      (global input) pointer to INTEGER
*          The global row index of the submatrix of the distributed
*          matrix A to operate on.
*
*  JA      (global input) pointer to INTEGER
*          The global column index of the submatrix of the distributed
*          matrix A to operate on.
*
*  DESCA   (global and local input) INTEGER array of dimension 8.
*          The array descriptor of the distributed matrix A.
*
*  BETA    (global input) pointer to REAL
*          On entry,  BETA  specifies the scalar beta.
*
*  C       (local input/local output) COMPLEX pointer into the
*          local memory to an array of dimension (LLD_C, LOCc(JC+N-1)).
*          Before entry with UPLO = 'U' or 'u', this array contains the
*          local pieces of the N-by-N upper triangular part of the
*          hermitian distributed matrix sub( C ) and its strictly lower
*          triangular part is not referenced. On exit, the upper trian-
*          gular part of sub( C ) is overwritten by the upper triangular
*          part of the updated distributed matrix.
*          Before entry  with  UPLO = 'L' or 'l', this array contains the
*          local pieces of the N-by-N lower triangular part of the
*          hermitian distributed matrix sub( C ) and its strictly upper
*          triangular part is not referenced. On exit, the lower trian-
*          gular part of sub( C ) is overwritten by the lower triangular
*          part of the updated distributed matrix.
*          Note that the imaginary parts of the diagonal elements of
*          sub( C ) need not be set, they are assumed to be zero, and
*          on exit they are set to zero.
*
*  IC      (global input) pointer to INTEGER
*          The global row index of the submatrix of the distributed
*          matrix C to operate on.
*
*  JC      (global input) pointer to INTEGER
*          The global column index of the submatrix of the distributed
*          matrix C to operate on.
*
*  DESCC   (global and local input) INTEGER array of dimension 8.
*          The array descriptor of the distributed matrix C.
*
*  =====================================================================
*
*  .. Local Scalars ..
*/
   char        * ctop, * rtop, matblk, TrA, UploA;
   int         block, i, iacol, iapos, iarow, iblk, iccol, icoffa,
               icoffc, icrow, ictxt, iia, iic, in, info, iroffa, iroffc,
               iszcmp, j, jblk, jja, jjc, jn, lcm, lcmp, lcmq, mullen,
               mycol, myrow, nca, ncc, notran, np0, nprow, npcol, nq0,
               nra, nrc, szcmp, tmp0, tmp1, tmp2, tmp3, wksz;
   float       tbeta;
/* ..
*  .. PBLAS Buffer ..
*/
   complex     * buff;
/* ..
*  .. External Functions ..
*/
   void        blacs_gridinfo_();
   void        pbchkmat();
   void        pberror_();
   char        * getpbbuf();
   char        * ptop();
   F_VOID_FCT  pbcherk_();
   F_INTG_FCT  ilcm_();
/* ..
*  .. Executable Statements ..
*
*  Get grid parameters
*/
   ictxt = desc_A[CTXT_];
   blacs_gridinfo_( &ictxt, &nprow, &npcol, &myrow, &mycol );
/*
*  Test the input parameters
*/
   info = 0;
   if( nprow == -1 )
      info = -(900+CTXT_+1);
   else
   {
      UploA = Mupcase( F2C_CHAR( uplo )[0] );
      TrA = Mupcase( F2C_CHAR( trans )[0] );
      notran =  ( TrA == 'N' );
      iroffa = (*ia-1) % desc_A[MB_];
      icoffa = (*ja-1) % desc_A[NB_];
      iroffc = (*ic-1) % desc_C[MB_];
      icoffc = (*jc-1) % desc_C[NB_];
      block = ( ( ( (*n) + iroffc ) <= desc_C[MB_] ) &&
                ( ( (*n) + icoffc ) <= desc_C[NB_] ) );
      if( notran )
      {
         pbchkmat( *n, 3, *k, 4, *ia, *ja, desc_A, 9, &iia, &jja,
                   &iarow, &iacol, nprow, npcol, myrow, mycol,
                   &nra, &nca, &info );
      }
      else
      {
         pbchkmat( *k, 4, *n, 3, *ia, *ja, desc_A, 9, &iia, &jja,
                   &iarow, &iacol, nprow, npcol, myrow, mycol,
                   &nra, &nca, &info );
      }
      pbchkmat( *n, 3, *n, 3, *ic, *jc, desc_C, 14, &iic, &jjc,
                &icrow, &iccol, nprow, npcol, myrow, mycol,
                &nrc, &ncc, &info );
      if( info == 0 )
      {
         if( ( UploA != 'U' ) && ( UploA != 'L' ) )
            info = -1;
         else if( ( TrA != 'N' ) && ( TrA != 'C' ) )
            info = -2;
         if( block )
         {
            if( notran )
            {
               if( ( nprow != 1 ) && ( (*n)+iroffa > desc_A[MB_] ) )
                  info = -7;
               else if( icrow != iarow )
                  info = -12;
            }
            else
            {
               if( ( npcol != 1 ) && ( (*n)+icoffa > desc_A[NB_] ) )
                  info = -8;
               else if( iccol != iacol )
                  info = -13;
            }
         }
         else
         {
            if( notran )
            {
               if( iroffa != 0 )
                  info = -7;
               else if( icrow != iarow )
                  info = -12;
               else if( desc_A[MB_] != desc_C[NB_] )
                  info = -(1400+MB_+1);
            }
            else
            {
               if( icoffa != 0 )
                  info = -8;
               else if( iccol != iacol )
                  info = -13;
               else if( desc_A[NB_] != desc_C[MB_] )
                  info = -(1400+NB_+1);
            }
            if( iroffc != 0 )
               info = -12;
            else if( icoffc != 0 )
               info = -13;
            else if( desc_C[MB_] != desc_C[NB_] )
               info = -(1400+NB_+1);
         }
         if( ictxt != desc_C[CTXT_] )
            info = -(1400+CTXT_+1);
      }
   }
   if( info )
   {
      pberror_( &ictxt, "PCHERK", &info );
      return;
   }
/*
*  Quick return if possible.
*/
   if( ( *n == 0 ) ||
       ( ( ( *alpha == ZERO ) || ( *k == 0 ) ) && ( *beta == ONE ) ) )
      return;
/*
*  Figure out the arguments to be passed to pbcherk
*/
   mullen = MULLENFAC * desc_C[MB_];
   if( block )
   {
      matblk = 'B';
      wksz = (*n) * (*n);
      if( notran )
         iapos = iacol;
      else
         iapos = iarow;
   }
   else
   {
      matblk = 'M';
      lcm = ilcm_( &nprow, &npcol );
      lcmp = lcm / nprow;
      lcmq = lcm / npcol;
      tmp3 = lcmq * desc_C[MB_];
      iszcmp = CEIL( mullen, tmp3 );
      szcmp = iszcmp * iszcmp * lcmq * desc_C[MB_] * lcmp * desc_C[MB_];
      tmp0 = CEIL( (*n), desc_C[NB_] );
      if( notran )
      {
          tmp1 = (*n) / desc_C[MB_];
          iapos = iacol;
          nq0 = MYROC0( tmp1, *n, desc_C[MB_], npcol );
          tmp2 = nq0 / desc_C[NB_];
          tmp3 = desc_A[NB_] * MYROC0( tmp2, nq0, desc_C[NB_], lcmq ) *
                 MIN( tmp0, lcmq );
          wksz = ( nq0 + MYROC0( tmp1, *n, desc_C[MB_], nprow ) ) *
                 desc_A[NB_] + MAX( szcmp, tmp3 );
      }
      else
      {
          tmp1 = (*n) / desc_C[MB_];
          iapos = iarow;
          np0 = MYROC0( tmp1, *n, desc_C[MB_], nprow );
          tmp2 = np0 / desc_C[MB_];
          tmp3 = desc_A[MB_] * MYROC0( tmp2, np0, desc_C[MB_], lcmp ) *
                 MIN( tmp0, lcmp );
          wksz = ( np0 + MYROC0( tmp1, *n, desc_C[MB_], npcol ) ) *
                 desc_A[MB_] + MAX( szcmp, tmp3 );
      }
   }
  buff = (complex *)getpbbuf( "PCHERK", wksz*sizeof(complex) );
/*
*  Call PB-BLAS routine
*/
   tbeta = ONE;
   if( block )
   {
      if( notran )
      {
         j = CEIL( (*ja), desc_A[NB_] ) * desc_A[NB_];
         jn = (*ja)+(*k)-1;
         jn = MIN( j, jn );
                                     /* Handle first block separately */
         jblk = jn-(*ja)+1;
         pbcherk_( &ictxt, C2F_CHAR( &matblk ), uplo, trans, n, &jblk,
                   &desc_A[NB_], alpha, &A[iia-1+(jja-1)*desc_A[LLD_]],
                   &desc_A[LLD_], beta, &C[iic-1+(jjc-1)*desc_C[LLD_]],
                   &desc_C[LLD_], &iapos, &icrow, &iccol,
                   C2F_CHAR( TOPDEF ), C2F_CHAR( NO ), C2F_CHAR( NO ),
                   &mullen, buff );
         if( mycol == iapos )
         {
            jja += jblk;
            jja = MIN( jja, nca );
         }
         iapos = (iapos+1) % npcol;
         jblk = (*k) - jblk;
         pbcherk_( &ictxt, C2F_CHAR( &matblk ), uplo, trans, n, &jblk,
                   &desc_A[NB_], alpha, &A[iia-1+(jja-1)*desc_A[LLD_]],
                   &desc_A[LLD_], &tbeta, &C[iic-1+(jjc-1)*desc_C[LLD_]],
                   &desc_C[LLD_], &iapos, &icrow, &iccol,
                   C2F_CHAR( TOPDEF ), C2F_CHAR( NO ), C2F_CHAR( NO ),
                   &mullen, buff );
      }
      else
      {
         i = CEIL( (*ia), desc_A[MB_] ) * desc_A[MB_];
         in = (*ia)+(*k)-1;
         in = MIN( i, in );
                                     /* Handle first block separately */
         iblk = in-(*ia)+1;
         pbcherk_( &ictxt, C2F_CHAR( &matblk ), uplo, trans, n, &iblk,
                   &desc_A[MB_], alpha, &A[iia-1+(jja-1)*desc_A[LLD_]],
                   &desc_A[LLD_], beta, &C[iic-1+(jjc-1)*desc_C[LLD_]],
                   &desc_C[LLD_], &iapos, &icrow, &iccol,
                   C2F_CHAR( TOPDEF ), C2F_CHAR( NO ), C2F_CHAR( NO ),
                   &mullen, buff );
         if( myrow == iapos )
         {
            iia += iblk;
            iia = MIN( iia, nra );
         }
         iapos = (iapos+1) % nprow;
         iblk = (*k) - iblk;
         pbcherk_( &ictxt, C2F_CHAR( &matblk ), uplo, trans, n, &iblk,
                   &desc_A[MB_], alpha, &A[iia-1+(jja-1)*desc_A[LLD_]],
                   &desc_A[LLD_], &tbeta, &C[iic-1+(jjc-1)*desc_C[LLD_]],
                   &desc_C[LLD_], &iapos, &icrow, &iccol,
                   C2F_CHAR( TOPDEF ), C2F_CHAR( NO ), C2F_CHAR( NO ),
                   &mullen, buff );
      }
   }
   else
   {
      if( notran )
      {
         rtop = ptop( BROADCAST, ROW, TOPGET );
         j = CEIL( (*ja), desc_A[NB_] ) * desc_A[NB_];
         jn = (*ja)+(*k)-1;
         jn = MIN( j, jn );
                                     /* Handle first block separately */
         jblk = jn-(*ja)+1;
         pbcherk_( &ictxt, C2F_CHAR( &matblk ), uplo, trans, n, &jblk,
                   &desc_A[MB_], alpha, &A[iia-1+(jja-1)*desc_A[LLD_]],
                   &desc_A[LLD_], beta, &C[iic-1+(jjc-1)*desc_C[LLD_]],
                   &desc_C[LLD_], &iapos, &icrow, &iccol,
                   C2F_CHAR( rtop ), C2F_CHAR( NO ), C2F_CHAR( NO ),
                   &mullen, buff );
         if( mycol == iapos )
         {
            jja += jblk;
            jja = MIN( jja, nca );
         }
         iapos = (iapos+1) % npcol;
                              /* loop over remaining block of columns */
         tmp0 = (*ja)+(*k)-1;
         for( j=jn+1; j<=tmp0; j+=desc_A[NB_] )
         {
             jblk = (*k)-j+(*ja);
             jblk = MIN( desc_A[NB_], jblk );
             pbcherk_( &ictxt, C2F_CHAR( &matblk ), uplo, trans, n, &jblk,
                       &desc_A[MB_], alpha, &A[iia-1+(jja-1)*desc_A[LLD_]],
                       &desc_A[LLD_], &tbeta, &C[iic-1+(jjc-1)*desc_C[LLD_]],
                       &desc_C[LLD_], &iapos, &icrow, &iccol,
                       C2F_CHAR( rtop ), C2F_CHAR( NO ), C2F_CHAR( NO ),
                       &mullen, buff );
             if( mycol == iapos )
             {
                jja += jblk;
                jja = MIN( jja, nca );
             }
             iapos = (iapos+1) % npcol;
         }
      }
      else
      {
         ctop = ptop( BROADCAST, COLUMN, TOPGET );
         i = CEIL( (*ia), desc_A[MB_] ) * desc_A[MB_];
         in = (*ia)+(*k)-1;
         in = MIN( i, in );
                                     /* Handle first block separately */
         iblk = in-(*ia)+1;
         pbcherk_( &ictxt, C2F_CHAR( &matblk ), uplo, trans, n, &iblk,
                   &desc_A[NB_], alpha, &A[iia-1+(jja-1)*desc_A[LLD_]],
                   &desc_A[LLD_], beta, &C[iic-1+(jjc-1)*desc_C[LLD_]],
                   &desc_C[LLD_], &iapos, &icrow, &iccol,
                   C2F_CHAR( ctop ), C2F_CHAR( NO ), C2F_CHAR( NO ),
                   &mullen, buff );
         if( myrow == iapos )
         {
            iia += iblk;
            iia = MIN( iia, nra );
         }
         iapos = (iapos+1) % nprow;
                                 /* loop over remaining block of rows */
         tmp0 = (*ia)+(*k)-1;
         for( i=in+1; i<=tmp0; i+=desc_A[MB_] )
         {
             iblk = *k-i+(*ia);
             iblk = MIN( desc_A[MB_], iblk );
             pbcherk_( &ictxt, C2F_CHAR( &matblk ), uplo, trans, n, &iblk,
                       &desc_A[NB_], alpha, &A[iia-1+(jja-1)*desc_A[LLD_]],
                       &desc_A[LLD_], &tbeta, &C[iic-1+(jjc-1)*desc_C[LLD_]],
                       &desc_C[LLD_], &iapos, &icrow, &iccol,
                       C2F_CHAR( ctop ), C2F_CHAR( NO ), C2F_CHAR( NO ),
                       &mullen, buff );
             if( myrow == iapos )
             {
                iia += iblk;
                iia = MIN( iia, nra );
             }
             iapos = (iapos+1) % nprow;
         }
      }
   }
}
